/*

pitchHeight~ - an FFT-based octave doubler.

Copyright 2009 William Brent

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <http://www.gnu.org/licenses/>.

version 0.0.5, January 20, 2011

 0.0.5 cleans up a few things, and rounds to the nearest bin index rather than just truncating  using floor().  I'm also killing the window/overlap functions, as people might as well only change that through creation arguments.  Also making the pitch input in MIDI units so that conversion isn't necessary.
0.0.4 got rid of a few unused variables. changed the ifdef NT to ifndef M_PI.
 0.0.3 adds a #define M_PI for windows compilation, and declares all functions except _setup static. also puts
pitchHeight_tilde *x = (pitchHeight_tilde *)pd_new(pitchHeight_tilde_class);
first in the new function.  

*/

#define NUMHARMONICS 25
#include "m_pd.h"
#include <math.h>
#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

static t_class *pitchHeight_tilde_class;

typedef struct _pitchHeight_tilde 
{
    t_object x_obj;
    float sr;
    float n;
    int dsp_tick;
    int buffer_limit;
    int overlap;
    int overlap_chunk;
	int pad;
    int window;
    int window_half;
	float ffa;
    float amp_scalar;
    float fundamental;
	float harmonics[NUMHARMONICS];
	int harmonicsBins[NUMHARMONICS];
    float fader;
    t_float *hann;
	t_sample *signal_buf;
	t_sample *signal_buf_windowed;
	t_float *signal_R;
	t_float *signal_I;
	
	t_sample *nonoverlapped_output;
	t_sample *final_output;
	
    float x_f;
    
} pitchHeight_tilde;


/* ------------------------ pitchHeight~ -------------------------------- */

static void pitchHeight_tilde_fader(pitchHeight_tilde *x, t_floatarg f)
{
	x->fader = 1-f;
}

static void pitchHeight_tilde_amp_scalar(pitchHeight_tilde *x, t_floatarg a)
{
	x->amp_scalar = a/x->window;
	
	post("amp scalar: %f", x->amp_scalar);
}

static void pitchHeight_tilde_pad(pitchHeight_tilde *x, t_floatarg p)
{
	x->pad = (p<0 ? 0 : p);
	x->pad = (x->pad>x->window_half ? x->window_half : x->pad);
	
	post("bin pad: %i", x->pad);
}

static void pitchHeight_tilde_fundamental(pitchHeight_tilde *x, t_floatarg fund)
{
    if (fund <= -1500)
    	x->fundamental = 0;
    else if (fund > 1499)
    	x->fundamental = (8.17579891564 * exp(.0577622650 * 1499));
    else
    	x->fundamental = (8.17579891564 * exp(0.0577622650 * fund));
}

static void *pitchHeight_tilde_new(t_floatarg window, t_floatarg overlap)
{
    pitchHeight_tilde *x = (pitchHeight_tilde *)pd_new(pitchHeight_tilde_class);
	int i, isPow2;
	
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("fundamental"));
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("fader"));
	outlet_new(&x->x_obj, gensym("signal"));

	if(window)
	{
		isPow2 = (int)window && !( ((int)window-1) & (int)window );
		if(isPow2)
			x->window = window;
		else
		{
			error("pitchHeight~: window size must be a power of 2.");
			x->window = 4096;
		}
	}
	else
		x->window = 4096;


	if(overlap)
	{
		isPow2 = (int)overlap && !( ((int)overlap-1) & (int)overlap );
		if(isPow2)
			x->overlap = overlap;
		else
		{
			error("pitchHeight~: overlap must be a power of 2.");
			x->overlap = 4;
		}
	}
	else
		x->overlap = 4;
		
	
	x->sr = 44100;
	x->n = 64;
	x->dsp_tick = 0; 
	x->window_half = x->window/2.0;
	x->overlap_chunk = x->window/x->overlap;
	x->amp_scalar = (2.0/3.0)/x->window;
	x->fundamental = 440.0;
	x->fader = 1;
	x->pad = floor(x->window * 0.00053257);
	x->ffa = x->sr/x->window;

	x->buffer_limit = x->window/x->n/x->overlap; // divide by overlap since we want to push out buffered audio to a window when the main buffer has been updated by 1/overlap the window sizse.
	
	x->hann = (t_float *)getbytes(x->window * sizeof(t_float));
	x->signal_buf = (t_sample *)getbytes(x->window * sizeof(t_sample));
	x->signal_buf_windowed = (t_sample *)getbytes(x->window * sizeof(t_sample));
	x->signal_R = (t_float *)getbytes((x->window_half+1) * sizeof(t_float));
	x->signal_I = (t_float *)getbytes((x->window_half+1) * sizeof(t_float));
	x->nonoverlapped_output = (t_sample *)getbytes((x->window*x->overlap) * sizeof(t_sample));
	x->final_output = (t_sample *)getbytes((x->overlap_chunk) * sizeof(t_sample));
	
	// initialize hann window
 	for(i=0; i<x->window; i++)
 		x->hann[i] = 0.5 * (1 - cos(2*M_PI*i/x->window));

 	for(i=0; i<x->window; i++)
 	{
 		x->signal_buf[i] = 0.0;
 		x->signal_buf_windowed[i] = 0.0;
	};

	// init output buffer
	for(i=0; i<(x->overlap_chunk); i++)
		x->final_output[i] = 0.0;
 			
	// initialize nonoverlapped_output
 	for(i=0; i<(x->window*x->overlap); i++)
 		x->nonoverlapped_output[i] = 0.0;
	
	// init (window_half+1 size arrays)
	for(i=0; i<=x->window_half; i++)
	{
		x->signal_R[i] = 0.0;
		x->signal_I[i] = 0.0;
	};	
	
    post("pitchHeight~: window size: %i, overlap: %i", x->window, x->overlap);
    
    return (x);
}


static t_int *pitchHeight_tilde_perform(t_int *w)
{
    int i, j, n, window, window_half, overlap, overlap_chunk;
    float amp_scalar;

    pitchHeight_tilde *x = (pitchHeight_tilde *)(w[1]);

    t_sample *in = (t_float *)(w[2]);
    t_sample *out = (t_float *)(w[3]);
    
    n = w[4];
	window = x->window;
	window_half = x->window_half;
	overlap = x->overlap;
	overlap_chunk = x->overlap_chunk;
	
	amp_scalar = x->amp_scalar;
	
	// shift previous contents back
	for(i=0; i<(window-n); i++)
		x->signal_buf[i] = x->signal_buf[n+i];
		
	// buffer most recent block
	for(i=0; i<n; i++)
		x->signal_buf[window-n+i] = in[i];
		
	
	if(x->dsp_tick>=x->buffer_limit)
	{
		x->dsp_tick = 0;
		
		// window the signal
		for(i=0; i<window; i++)
			x->signal_buf_windowed[i] = x->signal_buf[i] * x->hann[i];
	
		// take FT of window
		mayer_realfft(window, x->signal_buf_windowed);
	
		// unpack mayer_realfft results into R&I arrays
		for(i=0; i<=window_half; i++)
		{
			x->signal_R[i] = x->signal_buf_windowed[i];
			if(fabs(x->signal_R[i]) < 0.0001)
				x->signal_R[i] = 0.0;
		}
		
		x->signal_I[0]=0;  // DC
		
		for(i=(window-1), j=1; i>window_half; i--, j++)
		{
			x->signal_I[j] = x->signal_buf_windowed[i];
			if(fabs(x->signal_I[j]) < 0.0001)
				x->signal_I[j] = 0.0;
		}
		
		x->signal_I[window_half]=0;  // Nyquist

		// find harmonics for the current fundamental
		for(i=0; i<NUMHARMONICS; i++)
		{
			x->harmonics[i] = x->fundamental * (i+1);
			x->harmonicsBins[i] = floor(x->harmonics[i]/x->ffa + 0.5);
		}
		
		for(i=0; i<NUMHARMONICS; i=i+2)
		{
			int peak, rangeLo, rangeHi;
			
			peak = x->harmonicsBins[i];
			rangeLo = peak - x->pad;
			rangeHi = peak + x->pad;

			if(rangeLo < 0) rangeLo = 0;
			if(rangeHi >= window_half) rangeHi = window_half-1;

			for(j=rangeLo; j<=rangeHi; j++)
			{
				x->signal_R[j] *= x->fader;
				x->signal_I[j] *= x->fader;
			}
		}
		
	
		// pack real and imaginary parts in correct order for mayer_realifft
		for(i=0; i<=window_half; i++)
			x->signal_buf_windowed[i] = x->signal_R[i];
		
		for(i=(window_half+1), j=(window_half-1); i<window; i++, j--)
			x->signal_buf_windowed[i] = x->signal_I[j];
			
		// resynth
		mayer_realifft(window, x->signal_buf_windowed);
			
			
		// window
		for(i=0; i<window; i++)
		{
			x->signal_buf_windowed[i] *= x->hann[i];
			x->signal_buf_windowed[i] *= amp_scalar;
		}
			
		// shift nonoverlapped blocks back
		for(i=0; i<(overlap-1); i++)
			for(j=0; j<window; j++)
				x->nonoverlapped_output[(i*window)+j] = x->nonoverlapped_output[((i+1)*window)+j];
		
		// write the new block
		for(i=0; i<window; i++)
			x->nonoverlapped_output[((overlap-1)*window)+i] = x->signal_buf_windowed[i];
			
		// init this chunk of the final output so it can be summed in the for() below
		for(i=0; i<overlap_chunk; i++)
			x->final_output[i] = 0.0;
			
		// do the overlap/add
		for(i=0; i<overlap; i++)
			for(j=0; j<overlap_chunk; j++)
				x->final_output[j] += x->nonoverlapped_output[(i*window)+((overlap-i-1)*overlap_chunk)+j];
	};
	
	// output
	for(i=0; i<n; i++, out++)
		*out = x->final_output[(x->dsp_tick*n)+i];

	x->dsp_tick++;

    return (w+5);
}


static void pitchHeight_tilde_dsp(pitchHeight_tilde *x, t_signal **sp)
{
	dsp_add(
		pitchHeight_tilde_perform,
		4,
		x,
		sp[0]->s_vec,
		sp[1]->s_vec,
		sp[0]->s_n
	);

// compare n to stored n and recalculate hann & filterbank if different
	if( sp[0]->s_n != x->n  ||  sp[0]->s_sr != x->sr )
	{
		x->sr = sp[0]->s_sr;
		x->n = sp[0]->s_n;
		
		x->buffer_limit = x->window/x->n/x->overlap;
	};

};

static void pitchHeight_tilde_free(pitchHeight_tilde *x)
{	
    t_freebytes(x->hann, x->window*sizeof(t_float));
    t_freebytes(x->signal_buf, x->window*sizeof(t_sample));
    t_freebytes(x->signal_buf_windowed, x->window*sizeof(t_sample));
    t_freebytes(x->signal_R, (x->window_half+1)*sizeof(t_float));
    t_freebytes(x->signal_I, (x->window_half+1)*sizeof(t_float));
    t_freebytes(x->nonoverlapped_output, (x->window*x->overlap)*sizeof(t_sample));
    t_freebytes(x->final_output, (x->overlap_chunk)*sizeof(t_sample));
};

void pitchHeight_tilde_setup(void)
{
    pitchHeight_tilde_class = 
    class_new(
    	gensym("pitchHeight~"),
    	(t_newmethod)pitchHeight_tilde_new,
    	(t_method)pitchHeight_tilde_free,
        sizeof(pitchHeight_tilde),
        CLASS_DEFAULT,
        A_DEFFLOAT,
        A_DEFFLOAT,
		0
    );

    CLASS_MAINSIGNALIN(pitchHeight_tilde_class, pitchHeight_tilde, x_f);

	class_addmethod(
		pitchHeight_tilde_class, 
        (t_method)pitchHeight_tilde_fader,
		gensym("fader"),
		A_DEFFLOAT,
		0
	);

	class_addmethod(
		pitchHeight_tilde_class, 
        (t_method)pitchHeight_tilde_amp_scalar,
		gensym("amp_scalar"),
		A_DEFFLOAT,
		0
	);

	class_addmethod(
		pitchHeight_tilde_class, 
        (t_method)pitchHeight_tilde_pad,
		gensym("pad"),
		A_DEFFLOAT,
		0
	);
	
	class_addmethod(
		pitchHeight_tilde_class, 
        (t_method)pitchHeight_tilde_fundamental,
		gensym("fundamental"),
		A_DEFFLOAT,
		0
	);
	
    class_addmethod(
    	pitchHeight_tilde_class,
    	(t_method)pitchHeight_tilde_dsp,
    	gensym("dsp"),
    	0
    );
}
